################################################################################
#
# Copyright (C) 2023-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################
from typing import NamedTuple

def getCapabilitiesCache(rocmVersion: NamedTuple) -> dict:
    """Returns a compiler version dependent capabilities cache.

    Args:
        rocmVersion: The rocm major/minor version (often obtained from hipconfig maj/min version).

    Returns:
        A dictionary representing the capabilities cache.
    """

    # The base cachedAsmCaps is assumed to work for rocm versions <= 6.3
    # and amdclang major versions <= 18. The version dependent cachedAsmCaps
    # was introduced in ROCm 6.4 which contains clang 19.
    cachedAsmCaps = \
    {(0, 0, 0): {'HasAddLshl': False,
                 'HasAtomicAdd': False,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': False,
                 'HasExplicitCO': False,
                 'HasExplicitNC': False,
                 'HasGLCModifier': False,
                 'HasNTModifier': False,
                 'HasLshlOr': False,
                 'HasMFMA': False,
                 'HasMFMA_b8': False,
                 'HasMFMA_bf16_1k': False,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': False,
                 'HasMFMA_f64': False,
                 'HasMFMA_f8': False,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': False,
                 'HasMFMA_vgpr': False,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': False,
                 'HasWMMA': False,
                 'KernargPreloading': False,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 0,
                 'SupportedISA': False,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': False,
                 'v_dot2_f32_f16': False,
                 'v_dot2c_f32_f16': False,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': False,
                 'v_fma_f16': False,
                 'v_fma_f32': False,
                 'v_fma_f64': False,
                 'v_fma_mix_f32': False,
                 'v_fmac_f16': False,
                 'v_fmac_f32': False,
                 'v_mac_f16': False,
                 'v_mac_f32': False,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': False,
                 'v_pk_fma_f16': False,
                 'v_pk_fmac_f16': False},
     (8, 0, 3): {'HasAddLshl': False,
                 'HasAtomicAdd': False,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': False,
                 'HasExplicitNC': False,
                 'HasGLCModifier': True,
                 'HasNTModifier': False,
                 'HasLshlOr': False,
                 'HasMFMA': False,
                 'HasMFMA_b8': False,
                 'HasMFMA_bf16_1k': False,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': False,
                 'HasMFMA_f64': False,
                 'HasMFMA_f8': False,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': False,
                 'HasMFMA_vgpr': False,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': False,
                 'HasWMMA': False,
                 'KernargPreloading': False,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 15,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': False,
                 'v_dot2_f32_f16': False,
                 'v_dot2c_f32_f16': False,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': False,
                 'v_fma_f16': False,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': False,
                 'v_fmac_f16': False,
                 'v_fmac_f32': False,
                 'v_mac_f16': True,
                 'v_mac_f32': True,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': False,
                 'v_pk_fma_f16': False,
                 'v_pk_fmac_f16': False},
     (9, 0, 0): {'HasAddLshl': True,
                 'HasAtomicAdd': False,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': True,
                 'HasExplicitNC': False,
                 'HasGLCModifier': True,
                 'HasNTModifier': False,
                 'HasLshlOr': True,
                 'HasMFMA': False,
                 'HasMFMA_b8': False,
                 'HasMFMA_bf16_1k': False,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': False,
                 'HasMFMA_f64': False,
                 'HasMFMA_f8': False,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': False,
                 'HasMFMA_vgpr': False,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': True,
                 'HasWMMA': False,
                 'KernargPreloading': False,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 63,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': False,
                 'v_dot2_f32_f16': False,
                 'v_dot2c_f32_f16': False,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': False,
                 'v_fma_f16': True,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': False,
                 'v_fmac_f16': False,
                 'v_fmac_f32': False,
                 'v_mac_f16': True,
                 'v_mac_f32': True,
                 'v_mad_mix_f32': True,
                 'v_mov_b64': False,
                 'v_pk_fma_f16': True,
                 'v_pk_fmac_f16': False},
     (9, 0, 6): {'HasAddLshl': True,
                 'HasAtomicAdd': False,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': True,
                 'HasExplicitNC': False,
                 'HasGLCModifier': True,
                 'HasNTModifier': False,
                 'HasLshlOr': True,
                 'HasMFMA': False,
                 'HasMFMA_b8': False,
                 'HasMFMA_bf16_1k': False,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': False,
                 'HasMFMA_f64': False,
                 'HasMFMA_f8': False,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': False,
                 'HasMFMA_vgpr': False,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': True,
                 'HasWMMA': False,
                 'KernargPreloading': False,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 63,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': True,
                 'v_dot2_f32_f16': True,
                 'v_dot2c_f32_f16': False,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': False,
                 'v_fma_f16': True,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': True,
                 'v_fmac_f16': False,
                 'v_fmac_f32': True,
                 'v_mac_f16': True,
                 'v_mac_f32': True,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': False,
                 'v_pk_fma_f16': True,
                 'v_pk_fmac_f16': False},
     (9, 0, 8): {'HasAddLshl': True,
                 'HasAtomicAdd': True,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': True,
                 'HasExplicitNC': False,
                 'HasGLCModifier': True,
                 'HasNTModifier': False,
                 'HasLshlOr': True,
                 'HasMFMA': True,
                 'HasMFMA_b8': False,
                 'HasMFMA_bf16_1k': False,
                 'HasMFMA_bf16_original': True,
                 'HasMFMA_constSrc': False,
                 'HasMFMA_f64': False,
                 'HasMFMA_f8': False,
                 'HasMFMA_i8_908': True,
                 'HasMFMA_i8_940': False,
                 'HasMFMA_vgpr': False,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': True,
                 'HasWMMA': False,
                 'KernargPreloading': False,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 63,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': True,
                 'v_dot2_f32_f16': True,
                 'v_dot2c_f32_f16': True,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': True,
                 'v_fma_f16': True,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': True,
                 'v_fmac_f16': False,
                 'v_fmac_f32': True,
                 'v_mac_f16': True,
                 'v_mac_f32': True,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': False,
                 'v_pk_fma_f16': True,
                 'v_pk_fmac_f16': False},
     (9, 0, 10): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': False,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': True,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': True,
                  'HasMFMA_bf16_original': True,
                  'HasMFMA_constSrc': True,
                  'HasMFMA_f64': True,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': True,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': True,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': True,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': True,
                  'v_mac_f32': True,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (9, 4, 2): {'HasAddLshl': True,
                 'HasAtomicAdd': True,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': True,
                 'HasExplicitNC': False,
                 'HasGLCModifier': False,
                 'HasNTModifier': True,
                 'HasLshlOr': True,
                 'HasMFMA': True,
                 'HasMFMA_b8': True,
                 'HasMFMA_bf16_1k': True,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': True,
                 'HasMFMA_f64': True,
                 'HasMFMA_f8': True,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': True,
                 'HasMFMA_vgpr': True,
                 'HasMFMA_xf32': True,
                 'HasSMulHi': True,
                 'HasWMMA': False,
                 'KernargPreloading': True,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 63,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': True,
                 'v_dot2_f32_f16': True,
                 'v_dot2c_f32_f16': True,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': True,
                 'v_fma_f16': True,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': True,
                 'v_fmac_f16': False,
                 'v_fmac_f32': True,
                 'v_mac_f16': True,
                 'v_mac_f32': False,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': True,
                 'v_pk_fma_f16': True,
                 'v_pk_fmac_f16': False},
     (9, 5, 0): {'HasAddLshl': True,
                 'HasAtomicAdd': True,
                 'HasDirectToLdsDest': False,
                 'HasDirectToLdsNoDest': True,
                 'HasExplicitCO': True,
                 'HasExplicitNC': False,
                 'HasGLCModifier': False,
                 'HasNTModifier': True,
                 'HasLshlOr': True,
                 'HasMFMA': True,
                 'HasMFMA_b8': True,
                 'HasMFMA_bf16_1k': True,
                 'HasMFMA_bf16_original': False,
                 'HasMFMA_constSrc': True,
                 'HasMFMA_f64': True,
                 'HasMFMA_f8': True,
                 'HasMFMA_i8_908': False,
                 'HasMFMA_i8_940': True,
                 'HasMFMA_vgpr': True,
                 'HasMFMA_xf32': False,
                 'HasSMulHi': True,
                 'HasWMMA': False,
                 'KernargPreloading': True,
                 'MaxLgkmcnt': 15,
                 'MaxVmcnt': 63,
                 'SupportedISA': True,
                 'SupportedSource': True,
                 'VOP3v_dot4_i32_i8': True,
                 'v_dot2_f32_f16': True,
                 'v_dot2c_f32_f16': True,
                 'v_dot4_i32_i8': False,
                 'v_dot4c_i32_i8': True,
                 'v_fma_f16': True,
                 'v_fma_f32': True,
                 'v_fma_f64': True,
                 'v_fma_mix_f32': True,
                 'v_fmac_f16': False,
                 'v_fmac_f32': True,
                 'v_mac_f16': True,
                 'v_mac_f32': False,
                 'v_mad_mix_f32': False,
                 'v_mov_b64': True,
                 'v_pk_fma_f16': True,
                 'v_pk_fmac_f16': False},
     (10, 1, 0): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': False,
                  'v_dot2c_f32_f16': False,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': True,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (10, 1, 1): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': True,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (10, 1, 2): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': True,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (10, 3, 0): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (10, 3, 1): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (10, 3, 2): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': True,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': True,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': True,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (11, 0, 0): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': True,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (11, 0, 1): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': True,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (11, 0, 2): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': True,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (11, 0, 3): {'HasAddLshl': True,
              'HasAtomicAdd': True,
              'HasDirectToLdsDest': False,
              'HasDirectToLdsNoDest': False,
              'HasExplicitCO': True,
              'HasExplicitNC': True,
              'HasGLCModifier': True,
              'HasLshlOr': True,
              'HasMFMA': False,
              'HasMFMA_b8': False,
              'HasMFMA_bf16_1k': False,
              'HasMFMA_bf16_original': False,
              'HasMFMA_constSrc': False,
              'HasMFMA_f64': False,
              'HasMFMA_f8': False,
              'HasMFMA_i8_908': False,
              'HasMFMA_i8_940': False,
              'HasMFMA_vgpr': False,
              'HasMFMA_xf32': False,
              'HasNTModifier': False,
              'HasSMulHi': True,
              'HasWMMA': True,
              'KernargPreloading': False,
              'MaxLgkmcnt': 15,
              'MaxVmcnt': 63,
              'SupportedISA': True,
              'SupportedSource': True,
              'VOP3v_dot4_i32_i8': True,
              'v_dot2_f32_f16': True,
              'v_dot2c_f32_f16': True,
              'v_dot4_i32_i8': False,
              'v_dot4c_i32_i8': False,
              'v_fma_f16': True,
              'v_fma_f32': True,
              'v_fma_f64': True,
              'v_fma_mix_f32': True,
              'v_fmac_f16': False,
              'v_fmac_f32': True,
              'v_mac_f16': False,
              'v_mac_f32': False,
              'v_mad_mix_f32': False,
              'v_mov_b64': False,
              'v_pk_fma_f16': True,
              'v_pk_fmac_f16': False},
     (11, 5, 0): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': True,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (11, 5, 1): {'HasAddLshl': True,
                  'HasAtomicAdd': True,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': True,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': True,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': True,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (12, 0, 0): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': False,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': False,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False},
     (12, 0, 1): {'HasAddLshl': True,
                  'HasAtomicAdd': False,
                  'HasDirectToLdsDest': False,
                  'HasDirectToLdsNoDest': False,
                  'HasExplicitCO': True,
                  'HasExplicitNC': True,
                  'HasGLCModifier': False,
                  'HasNTModifier': False,
                  'HasLshlOr': True,
                  'HasMFMA': False,
                  'HasMFMA_b8': False,
                  'HasMFMA_bf16_1k': False,
                  'HasMFMA_bf16_original': False,
                  'HasMFMA_constSrc': False,
                  'HasMFMA_f64': False,
                  'HasMFMA_f8': False,
                  'HasMFMA_i8_908': False,
                  'HasMFMA_i8_940': False,
                  'HasMFMA_vgpr': False,
                  'HasMFMA_xf32': False,
                  'HasSMulHi': True,
                  'HasWMMA': False,
                  'KernargPreloading': False,
                  'MaxLgkmcnt': 15,
                  'MaxVmcnt': 63,
                  'SupportedISA': True,
                  'SupportedSource': True,
                  'VOP3v_dot4_i32_i8': False,
                  'v_dot2_f32_f16': True,
                  'v_dot2c_f32_f16': False,
                  'v_dot4_i32_i8': False,
                  'v_dot4c_i32_i8': False,
                  'v_fma_f16': True,
                  'v_fma_f32': True,
                  'v_fma_f64': True,
                  'v_fma_mix_f32': True,
                  'v_fmac_f16': False,
                  'v_fmac_f32': True,
                  'v_mac_f16': False,
                  'v_mac_f32': False,
                  'v_mad_mix_f32': False,
                  'v_mov_b64': False,
                  'v_pk_fma_f16': True,
                  'v_pk_fmac_f16': False}}

    def nested_update_(left, right):
        for key, val in right.items():
            left[key].update(val)

    if rocmVersion.major * 10 + rocmVersion.minor > 63:
        v19_diff = \
            {(11, 0, 0): {'VOP3v_dot4_i32_i8': True},
             (11, 0, 1): {'VOP3v_dot4_i32_i8': True},
             (11, 0, 2): {'VOP3v_dot4_i32_i8': True},
             (11, 0, 3): {'VOP3v_dot4_i32_i8': True},
             (11, 5, 0): {'VOP3v_dot4_i32_i8': True},
             (11, 5, 1): {'VOP3v_dot4_i32_i8': True},
             (12, 0, 0): {'VOP3v_dot4_i32_i8': True},
             (12, 0, 1): {'VOP3v_dot4_i32_i8': True}}

        nested_update_(cachedAsmCaps, v19_diff)

    return cachedAsmCaps



################################################################################
#
# Copyright (C) 2016-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from . import Common
from . import ClientWriter
from . import LibraryIO
from .Contractions import ProblemType as ContractionsProblemType
from .SolutionStructs import ProblemSizes, ProblemType
from .Common import tPrint, printExit, printWarning, assignGlobalParameters, \
        restoreDefaultGlobalParameters, HR
from .Tensile import addCommonArguments, argUpdatedGlobalParameters
from . import __version__

import argparse
import os
import sys


def getGlobalParams(config):
    """Try to parse out global parameters from the places it could be"""
    globalParams = None
    try:
        globalParams = config["GlobalParameters"]
    except (TypeError, LookupError):
        pass

    return globalParams


def getProblemDict(config):
    """Try to parse out a problem type dict from the places it could be"""
    problemDict = None
    try:  # Tensile Config file (first entry)
        problemDict = config["BenchmarkProblems"][0][0]
        if len(config["BenchmarkProblems"]) > 1:
            printWarning("More than one BenchmarkProblem in config file: only using first")
    except (TypeError, LookupError):
        pass
    else:
        return problemDict

    try:  # Solution Selection "base" file
        problemDict = config["ProblemType"]
    except (TypeError, LookupError):
        pass

    return problemDict


def getSizeList(config):
    """
    Try to parse out a size list from the places it could be
    (including the whole config being the list)
    """
    sizeList = None
    try:  # Tensile config file (first entry)
        sizeList = config["BenchmarkProblems"][0][1]["BenchmarkFinalParameters"][0]["ProblemSizes"]
    except (TypeError, LookupError):
        pass
    else:
        return sizeList

    # check if whole config looks like a size list
    keep = True
    if type(config) is list:
        for i in config:
            if not (type(i) is dict and ("Exact" in i or "Range" in i)):
                keep = False
                break
    else:
        keep = False

    if keep:
        sizeList = config
    return sizeList


def parseConfig(config):
    """Parse out all data we can get from the config"""
    globalParams = getGlobalParams(config)
    problemDict = getProblemDict(config)
    sizeList = getSizeList(config)

    return (globalParams, problemDict, sizeList)


def TensileClientConfig(userArgs):
    tPrint(1, "")
    tPrint(1, HR)
    tPrint(1, "#")
    tPrint(1, "#  Tensile Client Config v{}".format(__version__))
    tPrint(1, "#")
    tPrint(1, HR)
    tPrint(1, "")

    # argument parsing
    # yapf: disable
    argParser = argparse.ArgumentParser()
    argParser.add_argument("ConfigYaml", type=os.path.realpath, nargs="+",
            help="Config yaml(s) containing parameters and problem and size information")
    argParser.add_argument("--output-config", "-o", dest="OutputConfig", type=os.path.realpath,
            required=True, help="Path to output resulting client config file")
    argParser.add_argument("--merge-sizes", dest="MergeSizes", action="store_true",
            help="Allow sizes from multiple config files")
    # yapf: enable

    addCommonArguments(argParser)
    args = argParser.parse_args(userArgs)

    # loop through all configs and parse out all the data we can
    configs = [LibraryIO.readYAML(x) for x in args.ConfigYaml]

    globalParams = {}
    problemDict = None
    sizeList = None

    for config in configs:
        (myGlobalParams, myProblemDict, mySizeList) = parseConfig(config)

        # if we got data from the config, keep it
        # if we already had that data, warn and exit
        if myGlobalParams is not None:
            if globalParams == {} or globalParams == myGlobalParams:
                globalParams = myGlobalParams
            else:
                printExit("Multiple definitions for GlobalParameters found:\n{}\nand\n{}".format(
                    globalParams, myGlobalParams))

        if myProblemDict is not None:
            if problemDict is None or problemDict == myProblemDict:
                problemDict = myProblemDict
            else:
                printExit("Multiple definitions for ProblemType found:\n{}\nand\n{}".format(
                    problemDict, myProblemDict))

        if mySizeList is not None:
            if sizeList is None or sizeList == mySizeList:
                sizeList = mySizeList
            elif args.MergeSizes:
                sizeList += mySizeList
            else:
                printExit("Multiple size lists found:\n{}\nand\n{}\n"
                          "Run with --merge-sizes to keep all size lists found".format(
                              sizeList, mySizeList))

    if problemDict is None:
        printExit("No ProblemType found; cannot produce output")
    if sizeList is None:
        printExit("No SizeList found; cannot produce output")

    ssProblemType = ProblemType(problemDict)
    conProblemType = ContractionsProblemType.FromOriginalState(ssProblemType)
    sizes = ProblemSizes(ssProblemType, sizeList)  # TODO doesn't seem to work for range sizes

    # update globals
    restoreDefaultGlobalParameters()
    assignGlobalParameters(globalParams)

    overrideParameters = argUpdatedGlobalParameters(args)
    for key, value in overrideParameters.items():
        tPrint(1, "Overriding {0}={1}".format(key, value))
        Common.globalParameters[key] = value

    # write output
    ClientWriter.writeClientConfigIni(sizes, conProblemType, "", [], "", args.OutputConfig, None)


def main():
    TensileClientConfig(sys.argv[1:])



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import logging,pytest
from Tensile.SolutionStructs import ExactList
from Tensile.SolutionStructs import ExactDict
log =logging.getLogger("testlog")

@pytest.mark.parametrize("test_input, expected", [
    ([1,2,3,4], (1,2,3,4, 1,1,1,)),
    ([4,5,6,7], (4,5,6,7, 4,4,4)),
    ])
def test_exact_list(test_input, expected):
    z={
        "TotalIndices" : len(test_input),
        "OperationType" : 'GEMM',
        "NumIndicesC" : 2,
        "IndexAssignmentsLD" : [3,4,5,6],
        "NumIndicesLD" : 4,
        "IndexAssignmentsA" : [0,1,3],
        "IndexAssignmentsB" : [0,2,3],
        "ZeroPadA" : [],
        "ZeroPadB" : []
    }
    e = ExactList(test_input, z)
    assert(e.sizes == expected)

@pytest.mark.parametrize("test_input, expected", [
    ([1,2,3,4], (1,2,3,4, 1,1,1)),
    ([4,5,6,7], (4,5,6,7 ,4,4,4)),
    ([43,56,4,8], (43,56,4, 43,43,43,43)),
    ])
def test_exact_list_zp(test_input, expected):
    z={
        "TotalIndices" : len(test_input),
        "OperationType" : 'GEMM',
        "NumIndicesC" : 2,
        "IndexAssignmentsLD" : [3,4,5,6],
        "NumIndicesLD" : 4,
        "IndexAssignmentsA" : [0,1,3],
        "IndexAssignmentsB" : [0,2,3],
        "ZeroPadA" : [[1,4,5,7]],
        "ZeroPadB" : [[1,2,44,55]],
    }

    e = ExactList(test_input, z)
    assert(e.sizes == expected)
    assert(e.zeroPadA == [[1,4,5,7]])
    assert(e.zeroPadB == [[1,2,44,55]])


def test_exact_dict():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2,  3,4]],
        "ZeroPadB" : []
    }

    e = ExactDict({'sizes': [100,200,300]}, z)
    assert(e.sizes == [100,200,300])
    assert(e.stridesA == None)
    assert(e.zeroPadA == [[1,2,3,4]])
    assert(e.zeroPadB == [])

def test_exact_dict2():
    e = ExactDict({'sizes': [100,200,300], 'stridesA' : [-1,100,1000]}, None)
    assert(e.sizes == [100,200,300])
    assert(e.stridesA == [-1,100,1000])
    assert(e.zeroPadA == [])

def test_exact_padstart_left_tbd():
    z={
        "TotalIndices" : 3,
       "ZeroPadA" : [[1,2,  -1,-1]], # TBD pads should be specified at end:
       "ZeroPadB" : []
    }
    #with pytest.raises(RuntimeError, match="RuntimeError:.*"):
    with pytest.raises(RuntimeError):
        ExactDict({'sizes': [100,200,300], 'stridesA' : [-1,100,1000] }, z)

def test_exact_padstart_mismatch():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2, 3,4]],
        "ZeroPadB" : []
    }

    with pytest.raises(RuntimeError, match="problem-specified padStartA==6 does not match problem-type==3"):
        ExactDict({'sizes': [100,200,300], 'padStartA':[6]}, z)

def test_exact_padend_mismatch():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2, 3,4]],
        "ZeroPadB" : []
    }

    with pytest.raises(RuntimeError, match="problem-specified padEndA==7 does not match problem-type==4"):
        ExactDict({'sizes': [100,200,300], 'padEndA':[7]}, z)

def test_exact_dict3():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2, 6,4]],
        "ZeroPadB" : []
    }

    e = ExactDict({'sizes': [100,200,300], 'padStartA':[6]}, z)
    assert(e.sizes == [100,200,300])
    assert(e.zeroPadA == [[1,2, 6,4]])
    assert(e.stridesA == None)

def test_exact_override_pad_tbd():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2, -1,-1]],
        "ZeroPadB" : []
    }

    e = ExactDict({'sizes': [100,200,300], 'padStartA':[6], 'padEndA':[7]}, z)
    assert(e.sizes == [100,200,300])
    assert(e.zeroPadA == [[1,2, 6,7]])
    assert(e.stridesA == None)

def test_exact_override_pad_tbd_multi_pad():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [[1,2, -1,-1], [4,5, -1, -1]],
        "ZeroPadB" : []
    }

    e = ExactDict({'sizes': [100,200,300], 'padStartA':[6,66], 'padEndA':[7,77]}, z)
    assert(e.sizes == [100,200,300])
    assert(e.zeroPadA == [[1,2, 6,7], [4,5, 66,77]])
    assert(e.stridesA == None)

def test_exact_override_pad_tbd_multi_pad_b():
    z={
        "TotalIndices" : 3,
        "ZeroPadA" : [],
        "ZeroPadB" : [[1,2, -1,-1], [4,5, -1, -1]],
    }

    e = ExactDict({'sizes': [100,200,300], 'padStartB':[6,66], 'padEndB':[7,77]}, z)
    assert(e.sizes == [100,200,300])
    assert(e.zeroPadB == [[1,2, 6,7], [4,5, 66,77]])
    assert(e.stridesA == None)



################################################################################
#
# Copyright (C) 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from pathlib import Path
from Tensile.EmbeddedData import generateLibrary

referenceSource = \
"""/*******************************************************************************
* Copyright (C) 2016-2021 Advanced Micro Devices, Inc. All rights reserved.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell cop-
* ies of the Software, and to permit persons to whom the Software is furnished
* to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IM-
* PLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
* FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNE-
* CTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

/**************************************************
* This file was generated by Tensile:             *
* https://github.com/ROCmSoftwarePlatform/Tensile *
**************************************************/




#include <Tensile/EmbeddedData.hpp>

#include <Tensile/Contractions.hpp>
#include <Tensile/Tensile.hpp>

namespace Tensile {
    namespace {
        // myMasterLibrary
        EmbedData<MyBase> TENSILE_EMBED_SYMBOL_NAME("my-library-test", {
            0x31, 0x32, 0x33, 0x34, 0x00});
    } // anonymous namespace
    namespace {
        // mylib1.co
        EmbedData<SolutionAdapter> TENSILE_EMBED_SYMBOL_NAME("my-library-test", {
            0x35, 0x36, 0x37, 0x38});
    } // anonymous namespace
    namespace {
        // mylib2.co
        EmbedData<SolutionAdapter> TENSILE_EMBED_SYMBOL_NAME("my-library-test", {
            0x37, 0x38, 0x39, 0x30});
    } // anonymous namespace
} // namespace Tensile""".split("\n")

def test_generateLibrary():
  name = Path.cwd() / "my-library"
  key = "my-library-test"

  masterLibrary = Path.cwd() / "myMasterLibrary"
  with open(masterLibrary, "w") as f:
    f.write("1234")

  coFiles = [ Path.cwd() / "mylib1.co", Path.cwd() / "mylib2.co" ]
  data = ["5678", "7890"]
  for t in zip(coFiles, data):
    with open(t[0], "w") as f:
      f.write(t[1])

  generateLibrary(name, key, masterLibrary, "MyBase", coFiles)

  with open(name.with_suffix(".cpp")) as f:
    embedSource = f.readlines()
    for e, r in zip(embedSource, referenceSource):
      assert e.rstrip() == r, "Generated file does not match reference."



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from Tensile.Utilities.merge import cmpHelper, fixSizeInconsistencies, removeUnusedKernels, mergeLogic
from Tensile.Utilities.ConditionalImports import yamlLoader

import yaml, pytest

# the merge scripts does not differentiate solutions based on index or name
# so "DummyParam" is used to mark if solutions should be equal or not

logicPrefix = r"""
- DummyVersionRequirement
- DummyLanguage
- DummyScheduleName
- DummyDevice
- DummyProblemType
"""

baseLogic=logicPrefix + r"""
-
  - SolutionIndex: 2
    SolutionNameMin: InUseForSize256
    DummyParam: InUseForSize256
  - SolutionIndex: 1
    SolutionNameMin: UnusedSolution
    DummyParam: UnusedSolution
  - SolutionIndex: 0
    SolutionNameMin: InUseForSize128or64
    DummyParam: InUseForSize128or64
- DummyIndexAssignment
-
  - - [256, 256, 1, 256]
    - [2, 111.1]
  - - [128, 128, 1, 128]
    - [0, 99.9]
  - - [64, 64, 1, 64]
    - [0, 88.8]
"""

incLogic=logicPrefix + r"""
-
  - SolutionIndex: 39
    SolutionNameMin: InUseForSize256or1024xxx
    DummyParam: InUseForSize256or1024xxx
  - SolutionIndex: 1
    SolutionNameMin: InUseForSize128xxx
    DummyParam: InUseForSize128xxx
- DummyIndexAssignment
-
  - - [128, 128, 1, 128]
    - [1, 999.9]
  - - [256, 256, 1, 256]
    - [39, 999.9]
  - - [1024, 1024, 1, 1024]
    - [39, 999.9]
"""

notUniqueSolution=logicPrefix+r"""
-
  - SolutionIndex: 0
    SolutionNameMin: Kernel0
    DummyParam: Kernel0
  - SolutionIndex: 1
    SolutionNameMin: Kernel0
    DummyParam: Kernel0
"""

uniqueSolution=logicPrefix+r"""
-
  - SolutionIndex: 0
    SolutionNameMin: Kernel0
    DummyParam: Kernel0
  - SolutionIndex: 1
    SolutionNameMin: Kernel1
    DummyParam: Kernel1
"""

notTrimmedSize=r"""
-
  - - [128, 128, 1, 128, 128, 128, 128, 128]
    - [1, 999.9]
  - - [1024, 1024, 1, 1024]
    - [42, 0.0]
  - - [128, 128, 1, 128]
    - [39, 0.0]
  - - [1024, 1024, 1, 1024, 1024, 1024, 1024, 1024]
    - [39, 999.9]
  - - [1024, 1024, 1, 1024, 1044, 1044, 1044, 1044]
    - [39, 999.9]
"""

trimmedSize=r"""
-
  - - [128, 128, 1, 128]
    - [1, 999.9]
  - - [512, 512, 1, 512]
    - [39, 999.9]
  - - [1024, 1024, 1, 1024]
    - [42, 999.9]
"""

mfmaMergeBaseLogic=logicPrefix+r"""
-
  - SolutionIndex: 0
    SolutionNameMin: MFMA_base
    DummyParam: MFMA_base
    EnableMatrixInstruction: True
    MatrixInstruction: [16, 16, 4, 1]
  - SolutionIndex: 1
    SolutionNameMin: VALU_base
    DummyParam: VALU_base
    EnableMatrixInstruction: False
    MatrixInstruction: []
- DummyIndexAssignment
"""

mfmaMergeIncLogic=logicPrefix+r"""
-
  - SolutionIndex: 0
    SolutionNameMin: MFMA_inc
    DummyParam: MFMA_inc
    EnableMatrixInstruction: True
    MatrixInstruction: [16, 16, 4, 1]
  - SolutionIndex: 1
    SolutionNameMin: VALU_inc
    DummyParam: VALU_inc
    EnableMatrixInstruction: False
    MatrixInstruction: []
- DummyIndexAssignment
"""

mfmaMergeBaseSizes=r"""
-
  - - [128, 128, 1, 128]
    - [0, 3.0]
  - - [128, 128, 1, 128]
    - [1, 6.0]
  - - [130, 128, 1, 128]
    - [1, 9.0]
  - - [131, 128, 1, 128]
    - [0, 12.0]
"""

mfmaMergeIncFasterSizes=r"""
-
  - - [128, 128, 1, 128]
    - [0, 4.0]
  - - [128, 128, 1, 128]
    - [1, 7.0]
  - - [131, 128, 1, 128]
    - [0, 13.0]
  - - [130, 128, 1, 128]
    - [1, 10.0]
"""

mfmaMergeIncSlowerSizes=r"""
-
  - - [128, 128, 1, 128]
    - [0, 2.0]
  - - [128, 128, 1, 128]
    - [1, 5.0]
  - - [131, 128, 1, 128]
    - [0, 11.0]
  - - [130, 128, 1, 128]
    - [1, 8.0]
"""

mfmaMergeIncNotMatchingMFMA=r"""
-
  - - [130, 128, 1, 128]
    - [0, 7.0]
  - - [131, 128, 1, 128]
    - [1, 12.0]
"""

mfmaMergeResNotMatchingMFMA=r"""
-
  - - [128, 128, 1, 128]
    - [0, 3.0]
  - - [128, 128, 1, 128]
    - [1, 6.0]
  - - [130, 128, 1, 128]
    - [1, 9.0]
  - - [131, 128, 1, 128]
    - [0, 12.0]
  - - [130, 128, 1, 128]
    - [2, 7.0]
  - - [131, 128, 1, 128]
    - [3, 12.0]
"""

def checkUniqueSolution(solutionPool):
  uniq = set()
  # note: any([False or None, True or None]) -> True
  return not any(frozenset(cmpHelper(i).items()) in uniq or uniq.add(frozenset(cmpHelper(i).items()))
                 for i in solutionPool)

@pytest.mark.parametrize("sizes, expected", [
  (   trimmedSize, [[128,128,1,128],[512,512,1,512],[1024,1024,1,1024]]),
  (notTrimmedSize, [[1024,1024,1,1024],[128,128,1,128]])
  ])
def test_fixSizeInconsistencies(sizes, expected):
  data = yaml.load(sizes, yamlLoader)
  data_ = fixSizeInconsistencies(data[0], "dummy")

  for [size, [_,_]], expected_ in zip(data_[0], expected):
    assert size == expected_

@pytest.mark.parametrize("input,expectedNumKernelRemoved", [(baseLogic, 1),
                                                            (incLogic, 0)])
def test_removeUnusedKernels(input, expectedNumKernelRemoved):
  data = yaml.load(input, yamlLoader)
  dataFiltered, numKernelRemoved = removeUnusedKernels(data)

  # test if number of solution removed is correct
  assert numKernelRemoved == expectedNumKernelRemoved

  # test if solution is re-indexed
  for index, s in enumerate(dataFiltered[5]):
    assert index == s["SolutionIndex"]

  # check if size-kernel mapping is not compromised
  solutionMapBefore = {}
  solutionMapAfter = {}
  for [size, [index, _]] in data[7]:
    solutionMapBefore[tuple(size)] = [s["SolutionNameMin"] for s in data[5] if s["SolutionIndex"]==index][0]
  for [size, [index, _]] in dataFiltered[7]:
    solutionMapAfter[tuple(size)] = [s["SolutionNameMin"] for s in dataFiltered[5] if s["SolutionIndex"]==index][0]

  assert solutionMapBefore == solutionMapAfter

  # print final yaml for visual inspection
  # stream = io.StringIO("")
  # yaml.safe_dump(dataFiltered, stream, default_flow_style=None)
  # print(stream.getvalue())

@pytest.mark.parametrize("baseLogic, incLogic, expectedStats, expectedSizes, expectedSolutions", [
# test case #1: merge incLogic into baseLogic
  (baseLogic, incLogic, [1,2,2], # 1 sizes added, 2 kernels added, 2 kernels removed/replaced
  [(1024,1024,1,1024), (256,256,1,256), (128,128,1,128), (64,64,1,64)],
  ["InUseForSize256or1024xxx", "InUseForSize256or1024xxx", "InUseForSize128xxx", "InUseForSize128or64"]),
# test case #2: merge baseLogic into itself
  (baseLogic, baseLogic, [0,0,1], # 0 sizes added, 0 kernels added, 1 kernel removed because it's unused
  [(256,256,1,256), (128,128,1,128), (64,64,1,64)],
  ["InUseForSize256", "InUseForSize128or64", "InUseForSize128or64"]),
])
def test_mergeLogic(baseLogic, incLogic, expectedStats, expectedSizes, expectedSolutions):
  baseData = yaml.load(baseLogic, yamlLoader)
  incData = yaml.load(incLogic, yamlLoader)

  mergedData, *stats = mergeLogic(baseData, incData, False)

  # check if stats are as expected
  assert stats == expectedStats

  # check if solution matches expected. assumes SolutionNameMin is uniqueSolution
  # (which is satisfied by the test data given here)
  solutionMap = {} # size -> solutionName
  for size, [index, _] in mergedData[7]:
    solutionMap[tuple(size)] = [s["SolutionNameMin"] for s in mergedData[5] if s["SolutionIndex"]==index][0]

  for size, expected in zip(expectedSizes, expectedSolutions):
    assert solutionMap[size] == expected

  # check if each solution in merged data is uniqueSolution
  assert checkUniqueSolution(mergedData[5])

@pytest.mark.parametrize("input,expected", [(uniqueSolution, True), (notUniqueSolution, False)])
def test_checkUniqueSolution(input, expected):
  data = yaml.load(input, yamlLoader)
  assert checkUniqueSolution(data[5]) == expected

@pytest.mark.parametrize("baseLogic, incLogic, expectedSizesYaml, expectedSolutions", [
# test case #1: Slower sizes in incremental logic file
  (mfmaMergeBaseLogic+mfmaMergeBaseSizes, mfmaMergeIncLogic+mfmaMergeIncSlowerSizes,
   mfmaMergeBaseSizes, ["MFMA_base", "VALU_base"]),
# test case #2: Faster sizes in incremental logic file
  (mfmaMergeBaseLogic+mfmaMergeBaseSizes, mfmaMergeIncLogic+mfmaMergeIncFasterSizes,
   mfmaMergeIncFasterSizes, ["MFMA_inc", "VALU_inc"]),
# test case #3: Test that VALU size is included alongside MFMA size, and vice versa (regardless of efficiency)
  (mfmaMergeBaseLogic+mfmaMergeBaseSizes, mfmaMergeIncLogic+mfmaMergeIncNotMatchingMFMA,
   mfmaMergeResNotMatchingMFMA, ["MFMA_base", "VALU_base", "MFMA_inc", "VALU_inc"])
])
def test_mfmaMergeLogic(baseLogic, incLogic, expectedSizesYaml, expectedSolutions):
  baseData      = yaml.load(baseLogic, yamlLoader)
  incData       = yaml.load(incLogic,  yamlLoader)
  expectedSizes = yaml.load(expectedSizesYaml, yamlLoader)[0]

  mergedData, _, _, _ = mergeLogic(baseData, incData, False, True, True)

  solutionIndices = {s['SolutionNameMin']: s['SolutionIndex'] for s in mergedData[5]} # size -> solutionName

  #Ensure all correct solutions are present in merged data
  for solution in expectedSolutions:
    assert solution in solutionIndices.keys()

  assert len(expectedSolutions) == len(mergedData[5])

  #Convert expected sizes to use mergedData's solution indices
  expectedSizes = [ [size, [solutionIndices[expectedSolutions[solIndex]], eff]] for size, [solIndex, eff] in expectedSizes ]

  #Ensure all expected sizes are present in merged data
  for item in expectedSizes:
    assert item in mergedData[7]

  assert len(expectedSizes) == len(mergedData[7])

if __name__ == "__main__":
    # test_mergeLogic(baseLogic, incLogic, [1,2,2], [(1024, 1024, 1, 1024), (256, 256, 1, 256), (128, 128, 1, 128), (64, 64, 1, 64)], [ "InUseForSize256or1024xxx", "InUseForSize256or1024xxx", "InUseForSize128xxx", "InUseForSize128or64"])
    # test_checkUniqueSolution(uniqueSolution, True)
    # test_fixSizeInconsistencies(trimmedSize, [[128,128,1,128],[512,512,1,512],[1024,1024,1,1024]])
    test_fixSizeInconsistencies(notTrimmedSize, [[1024,1024,1,1024],[128,128,1,128]])



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import pytest

import Tensile.Component  as Component
import Tensile.Components as Components
from Tensile.DataType import DataType

def test_PartialMatch():
    a = {'foo': True,
         'bar': 25,
         'baz': {'Enabled': True,
                 'Debug': False}}

    b = {'foo': True}

    assert Component.PartialMatch(b, a)
    assert not Component.PartialMatch(a, b)

    assert not Component.PartialMatch({'foo': False}, a)

    assert not Component.PartialMatch({'baz': {"Enabled": False}}, a)
    assert Component.PartialMatch({'baz': {"Enabled": True}},  a)

    assert not Component.PartialMatch({'baz': {"Error": True}}, a)

    assert not Component.PartialMatch({'bar': lambda x: x < 10}, a)
    assert Component.PartialMatch({'bar': lambda x: x > 10}, a)

    shouldMatch = lambda obj: obj['foo'] and obj['bar'] > 20
    shouldNotMatch = lambda obj: obj['foo'] and obj['bar'] < 20

    assert Component.PartialMatch(shouldMatch, a)
    assert not Component.PartialMatch(shouldNotMatch, a)

class MockWriter:
    def __init__(self, **kwargs):
        defaultArgs = {'endLine': '\n'}

        args = {}
        args.update(defaultArgs)
        args.update(kwargs)

        for k,v in args.items():
            setattr(self, k, v)

@pytest.fixture
def vega10():
    return {
        'asmCaps': {'v_fma_f16': False,
                   'v_pk_fma_f16': True,
                   'v_dot2c_f32_f16': False,
                   'v_dot2_f32_f16': False,
                   'v_dot4c_i32_i8': False,
                   'v_dot4_i32_i8': False,
                   "v_mad_mix_f32": True,
                   "v_fma_mix_f32": False,
                   "v_mac_f32": True,
                   "v_fma_f32": True,
                   "v_fmac_f32": False,
                   }
    }

@pytest.fixture
def navi10():
    return {
        'asmCaps': {'v_fma_f16': True,
                   'v_pk_fma_f16': False,
                   'v_dot2c_f32_f16': False,
                   'v_dot2_f32_f16': False,
                   'v_dot4c_i32_i8': False,
                   'v_dot4_i32_i8': False,
                   "v_mad_mix_f32": False,
                   "v_fma_mix_f32": True,
                   "v_mac_f32": True,
                   "v_fma_f32": True,
                   "v_fmac_f32": True}
    }

@pytest.fixture
def navi12():
    return {
        'asmCaps': {'v_fma_f16': False,
                   'v_pk_fma_f16': False,
                   'v_dot2c_f32_f16': True,
                   'v_dot2_f32_f16': True,
                   'v_dot4c_i32_i8': True,
                   'v_dot4_i32_i8': True,
                   "v_mad_mix_f32": False,
                   "v_fma_mix_f32": True,
                   "v_mac_f32": True,
                   "v_fma_f32": True,
                   "v_fmac_f32": True}
    }

@pytest.fixture
def f16():
    return {
        'kernel': {"ProblemType": {"DataType": DataType(DataType.half),
                                   "HighPrecisionAccumulate": False},
                   "AggressivePerfMode": True,
                   "LocalDotLayout": 1,
                   "InnerUnroll": 1,
                   "ThreadTile0": 4,
                   "ThreadTile1": 4}
    }

@pytest.fixture
def f16_hpa():
    return {
        'kernel': {"ProblemType": {"DataType": DataType(DataType.half),
                                   "HighPrecisionAccumulate": True},
                   "AggressivePerfMode": True,
                   "LocalDotLayout": 1,
                   "InnerUnroll": 1,
                   "ThreadTile0": 4,
                   "ThreadTile1": 4}
    }

@pytest.fixture
def f16_hpa_ldl():
    return {
        'kernel': {"ProblemType": {"DataType": DataType(DataType.half),
                                   "HighPrecisionAccumulate": True},
                   "AggressivePerfMode": True,
                   "LocalDotLayout": 2,
                   "InnerUnroll": 2,
                   "ThreadTile0": 4,
                   "ThreadTile1": 4}
    }

#navi = MockWriter(asmCaps = {'v_fma_f16': True,
#                             'v_pk_fma_f16': False},
#                  kernel = {"ProblemType": {"DataType": DataType(DataType.half),
#                                            "HighPrecisionAccumulate": False},
#                            "AggressivePerfMode": True,
#                            "ThreadTile0": 4,
#                            "ThreadTile1": 4},
#                  endLine = '\n')

def test_find(navi10, f16):
    writer = MockWriter(**navi10, **f16)

    found = Component.MAC.find(writer)
    assert isinstance(found, Components.MAC_F16.FMA_F16_NonPacked)

def test_find2(vega10, f16_hpa):
    writer = MockWriter(**vega10, **f16_hpa)

    found = Component.MAC.find(writer)
    assert isinstance(found, Components.MAC_F16_HPA.FMA_F16_HPA_MAD_MIX)

def test_MAC_F16_FMA_NonPacked(navi10, f16):
    writer = MockWriter(**navi10, **f16)

    found = Component.MAC.find(writer)
    kernelText = found(writer, 2, 4)
    print(kernelText)

def test_componentPath():
    assert Components.MAC_F16.FMA_F16_NonPacked.componentPath() == ["Component", "MAC", "FMA_F16_NonPacked"]

def test_find_macs(useGlobalParameters, f16, f16_hpa, f16_hpa_ldl):
    with useGlobalParameters() as globals:
        for dtype in [f16, f16_hpa, f16_hpa_ldl]:
            for arch in globals["SupportedISA"]:
                writer = MockWriter(asmCaps=globals["AsmCaps"][arch], archCaps=globals["ArchCaps"][arch], **dtype)

                found = Component.MAC.find(writer, True)
                # No HPA on 803, every other combination should work though.
                if arch != (8,0,3) or (dtype != f16_hpa and dtype != f16_hpa_ldl):
                    assert isinstance(found, Component.MAC)
                print(dtype, arch, found)



################################################################################
#
# Copyright (C) 2019-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

import functools
import sys
import time

from typing import List

from .Common import DeveloperWarning, printWarning

class ProgressBar:
    """A class for displaying a progress bar in the console.

    This class provides a simple way to display and update a progress bar in the console
    to indicate the progress of a long-running operation. The progress bar can be updated
    incrementally and supports displaying a completion message and time upon finishing.

    Attributes:
        char: The character used to fill the progress bar.
        maxValue: The maximum value the progress bar can represent.
        width: The total width of the progress bar, including borders.
        maxTicks: The maximum number of ticks (fill characters) within the progress bar.
        priorValue: The value of the progress bar during the last update.
        fraction: The fraction of the progress bar that is filled.
        numTicks: The current number of ticks (fill characters) in the progress bar.
        createTime: The timestamp when the progress bar was created.
        message: The message displayed alongside the progress bar.
    """
    def __init__(self, maxValue: int, desc: str, width=40):
        self.char: str = '.'
        self.maxValue: int = maxValue
        self.width: int = width
        self.maxTicks: int = self.width - 10  # Adjusted for better alignment

        self.priorValue: int = 0
        self.fraction: float = 0
        self.numTicks: int = 0
        self.createTime: float = time.time()

        self.message: str = "# " + desc

    def increment(self, value=1):
        """Increments the progress bar by a given value and updates the display."""
        self.update(self.priorValue + value)

    def update(self, value):
        """Updates the progress bar to a specific value and refreshes the display."""
        currentFraction = 1.0 * value / self.maxValue
        currentNumTicks = int(currentFraction * self.maxTicks)
        if currentNumTicks > self.numTicks:
            self.numTicks = currentNumTicks
            self.fraction = currentFraction
            self.printStatus()
        self.priorValue = value

    def printStatus(self):
        """Prints the current status of the progress bar to the console."""
        progress_bar = self.char * self.numTicks + ' ' * (self.maxTicks - self.numTicks)
        status_msg = f"{self.message} {progress_bar} {self.fraction * 100:.1f}%"

        if self.numTicks == 0:
            sys.stdout.write(status_msg)
        else:
            sys.stdout.write('\r' + ' ' * len(status_msg) + '\r' + status_msg)
        sys.stdout.flush()

    def finish(self):
        """Marks the operation as done, cleans up the display, and prints the completion time."""
        stopTime = time.time()

        sys.stdout.write(f" (took {stopTime - self.createTime:.1f} secs)\n")
        sys.stdout.flush()

class SpinnyThing:
    """A class to display a spinning indicator in the console for long-running operations.

    This class provides a simple way to visually indicate that a long-running operation
    is in progress by displaying a spinning character in the console. The spinner is
    updated at regular intervals, and a completion message with the elapsed time is
    displayed when the operation finishes.

    Attributes:
        msg: The message displayed alongside the spinner.
        chars: The sequence of characters used for the spinner animation.
        index: The current index in the `chars` list for the spinner.
        count: A counter to control the update frequency of the spinner.
        createTime: The timestamp when the spinner was created.
    """
    def __init__(self, desc: str):
        self.message: str = "# " + desc
        self.chars: List[str] = ['|', '/', '-', '\\']
        self.index: int = 0
        self.count: int = 0
        self.createTime: float = time.time()

    def increment(self):
        """Increments the spinner's position and updates the display if necessary."""
        self.count += 1
        if self.count % 3 != 0:
            return

        sys.stdout.write('\r' + ' ' * (len(self.message) + 10))
        sys.stdout.flush()

        sys.stdout.write('\r' + self.message + " " + self.chars[self.index])
        sys.stdout.flush()
        self.index = (self.index + 1) % len(self.chars)

    def finish(self):
        """Clears the spinner and displays a completion message with the elapsed time."""
        sys.stdout.write('\r' + ' ' * (len(self.message) + 10))
        sys.stdout.flush()

        stopTime = time.time()
        elapsedTime = stopTime - self.createTime

        sys.stdout.write('\r' + self.message + f'... Done in {elapsedTime:.1f} secs\n')
        sys.stdout.flush()

def iterate_progress(obj, *args, **kwargs):
    if 'desc' not in kwargs:
        printWarning("No message provided for TQDM progress bar/spinner", DeveloperWarning)
        kwargs['desc'] = 'Processing unknown function'
    if 'total' in kwargs:
        progress = ProgressBar(kwargs['total'], kwargs['desc'])
    else:
        try:
            progress =  ProgressBar(len(obj), kwargs['desc'])
        except TypeError:
            progress = SpinnyThing(kwargs['desc'])
    for o in obj:
        yield o
        progress.increment()
    progress.finish()

try:
    from tqdm import tqdm
except ImportError:
    tqdm = iterate_progress

def state(obj):
    if hasattr(obj, 'state'):
        return obj.state()

    if hasattr(obj.__class__, 'StateKeys'):
        rv = {}
        for key in obj.__class__.StateKeys:
            attr = key
            if isinstance(key, tuple):
                (key, attr) = key
            rv[key] = state(getattr(obj, attr))
        return rv

    if isinstance(obj, dict):
        return dict([(k, state(v)) for k,v in list(obj.items())])

    if any([isinstance(obj, cls) for cls in [str, int, float]]):
        return obj

    try:
        obj = [state(i) for i in obj]
        return obj
    except TypeError:
        pass

    return obj

def state_key_ordering(cls):
    def tup(obj):
        return tuple([getattr(obj, k) for k in cls.StateKeys])

    def lt(a, b):
        return tup(a) < tup(b)
    def eq(a, b):
        return tup(a) == tup(b)

    cls.__lt__ = lt
    cls.__eq__ = eq

    return functools.total_ordering(cls)

def hash_combine(*objs, **kwargs):
    shift = 1
    if 'shift' in kwargs:
        shift = kwargs['shift']

    if len(objs) == 1:
        objs = objs[0]

    rv = 0
    try:
        it = iter(objs)
        rv = next(it)
        for value in it:
            rv = (rv << shift) ^ value
    except TypeError:
        return objs
    except StopIteration:
        pass
    return rv

def hash_objs(*objs, **kwargs):
    return hash(tuple(objs))

def ceil_divide(numerator, denominator):
    # import pdb
    # pdb.set_trace()
    try:
        if numerator < 0 or denominator < 0:
            raise ValueError
    except ValueError:
        print("ERROR: Can't have a negative register value")
        return 0
    try:
        div = int((numerator+denominator-1) // denominator)
    except ZeroDivisionError:
        print("ERROR: Divide by 0")
        return 0
    return div

def roundUpToNearestMultiple(numerator, denominator):
    return ceil_divide(numerator,denominator)*int(denominator)



################################################################################
#
# Copyright (C) 2016-2023 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from . import Common
from . import LibraryIO
from .Common import globalParameters, pushWorkingPath, popWorkingPath, tPrint, printExit, printWarning, ClientExecutionLock
from .SolutionStructs import ProblemType, ProblemSizesMock
from .TensileCreateLibrary import copyStaticFiles

import os
import subprocess
import shutil
from enum import Enum

from .Contractions import FreeIndex
from .Contractions import ProblemType as ContractionsProblemType


class DataInitName(Enum):
  Zero = 0
  One = 1
  Two = 2
  Random = 3
  NaN = 4
  Inf = 5
  BadInput = 6
  BadOutput = 7
  SerialIdx = 8
  SerialDim0 = 9
  SerialDim1 = 10
  Identity = 11
  TrigSin = 12
  TrigCos = 13
  TrigAbsSin = 14
  TrigAbsCos = 15
  RandomNarrow = 16
  NegOne = 17
  Max = 18
  DenormMin = 19
  DenormMax = 20

class ClientLogLevel(Enum):
  Error = 0
  Terse = 1
  Verbose = 2
  Debug = 3


################################################################################
# Main
################################################################################
def main( config ):
  libraryLogicPath = os.path.join(globalParameters["WorkingPath"], globalParameters["LibraryLogicPath"])
  stepBaseDir = pushWorkingPath(globalParameters["LibraryClientPath"])

  pushWorkingPath("source")
  copyStaticFiles()

  ##############################################################################
  # Read Logic Files
  ##############################################################################
  logicFiles = [os.path.join(libraryLogicPath, f) for f \
      in os.listdir(libraryLogicPath) \
      if (os.path.isfile(os.path.join(libraryLogicPath, f)) \
      and os.path.splitext(f)[1]==".yaml")]
  tPrint(1, "LogicFiles: %s" % logicFiles)
  functions = []
  functionNames = []

  createLibraryScript = getBuildClientLibraryScript(stepBaseDir, libraryLogicPath)
  subprocess.run(createLibraryScript, cwd=stepBaseDir)
  coList = []
  yamlList = []

  with open(os.path.join(stepBaseDir,"library","TensileManifest.txt"), "r") as f:
    lines = f.read().split("\n")
    coList = [line for line in lines if "co" in line]
    yamlList = [line for line in lines if (("yaml" in line) or ("dat" in line))]

  clientParametersPaths = []
  for logicFileName in logicFiles:
    (scheduleName, _, problemType, _, exactLogic, newLibrary) \
        = LibraryIO.parseLibraryLogicFile(logicFileName)
    functions.append((scheduleName, problemType))
    functionNames.append("tensile_%s" % (problemType))
    problemSizes = ProblemSizesMock(exactLogic)
    clientParametersPaths.append(writeClientConfig(
                                  forBenchmark=False,
                                  solutions=None,
                                  problemSizes=problemSizes,
                                  stepName=str(ProblemType(problemType)),
                                  stepBaseDir=globalParameters["WorkingPath"],
                                  newLibrary=newLibrary,
                                  configBase="ClientParameters_%s"%str(ProblemType(problemType)),
                                  codeObjectFiles=coList,
                                  tileAwareSelection=False,
                                  libraryFile=yamlList[0]))

  ##############################################################################
  # Write Generated Header
  ##############################################################################
  forBenchmark = False
  problemSizes = None
  popWorkingPath() # source

  ##############################################################################
  # Run Build Script
  ##############################################################################
  # if redo=true, clobber the build directory
  if globalParameters["ForceRedoLibraryClient"]:
    shutil.rmtree(os.path.join(globalParameters["WorkingPath"], "build"), \
        ignore_errors=True)

  forBenchmark = False
  enableTileSelection = False
  returncode = runClient(libraryLogicPath, forBenchmark, enableTileSelection, clientParametersPaths)

  popWorkingPath() # LibraryClient

  return returncode

################################################################################
# Write Run Script
################################################################################
def runNewClient(scriptPath, clientParametersPath, clientBuildDir=None):

  clientExe = getClientExecutablePath()

  iniFile = "--config-file={}".format(clientParametersPath)
  args = [clientExe, iniFile]

  try:
    subprocess.run(args, check=True)
  except (subprocess.CalledProcessError, OSError) as e:
    printWarning("ClientWriter Benchmark Process exited with error: {}".format(e))


def runClient(libraryLogicPath, forBenchmark, enableTileSelection, configPaths=None):
  # write runScript
  pushWorkingPath("build")
  path = globalParameters["WorkingPath"]

  runScriptName = writeRunScript(path, forBenchmark, enableTileSelection, configPaths)
  with ClientExecutionLock():
    process = subprocess.Popen(runScriptName, cwd=path)
    process.communicate()

  if process.returncode:
    printWarning("ClientWriter Benchmark Process exited with code %u" % process.returncode)
  popWorkingPath() # build

  return process.returncode


def getBuildClientLibraryScript(buildPath, libraryLogicPath):
  callCreateLibraryCmd = ["python"] if os.name == "nt" else []

  callCreateLibraryCmd += [os.path.join(globalParameters["ScriptPath"] , "bin", "TensileCreateLibrary")]

  if globalParameters["MergeFiles"]:
    callCreateLibraryCmd += ["--merge-files"]
  else:
    callCreateLibraryCmd += ["--no-merge-files"]

  if globalParameters["ShortNames"]:
    callCreateLibraryCmd += ["--short-file-names"]
  else:
    callCreateLibraryCmd += ["--no-short-file-names"]

  if globalParameters["GenerateManifestAndExit"]:
    callCreateLibraryCmd += ["--generate-manifest-and-exit"]

  callCreateLibraryCmd += ["--architecture=" + globalParameters["Architecture"]]
  callCreateLibraryCmd += ["--code-object-version=" + globalParameters["CodeObjectVersion"]]
  callCreateLibraryCmd += ["--cxx-compiler=" + globalParameters["CxxCompiler"]]
  callCreateLibraryCmd += ["--library-format=" + globalParameters["LibraryFormat"]]

  callCreateLibraryCmd += ["%s" % libraryLogicPath]
  callCreateLibraryCmd += ["%s" % buildPath] #" ../source"
  callCreateLibraryCmd += ["%s" % globalParameters["RuntimeLanguage"]]

  return callCreateLibraryCmd


def writeRunScript(path, forBenchmark, enableTileSelection, configPaths=None):
  if configPaths is None:
    configPaths = []
    configPaths.append(os.path.join(globalParameters["WorkingPath"], "../source/ClientParameters.ini"))
    if enableTileSelection is True:
      configPaths.append(os.path.join(globalParameters["WorkingPath"], "../source/ClientParameters_Granularity.ini"))

  # create run.bat or run.sh which builds and runs
  clientExe = getClientExecutablePath()

  runScriptName = os.path.join(path, "run.%s" % ("bat" if os.name == "nt" else "sh") )
  runScriptFile = open(runScriptName, "w")
  if os.name != "nt":
    runScriptFile.write("#!/bin/bash\n\n")

  option = "" if forBenchmark else "--best-solution 1"
  if (os.name == "nt"):
    runScriptFile.write("@echo off\n")
    runScriptFile.write("set err=0\n")
    for configFile in configPaths:
      runScriptFile.write("{} --config-file {} {} {}\n".format(clientExe, configFile, globalParameters["ClientArgs"], option))
      runScriptFile.write("IF %errorlevel% NEQ 0 set err=%errorlevel%\n")
    runScriptFile.write("exit %err%\n")
  else:
    runScriptFile.write("set -ex\n")
    if globalParameters["PinClocks"] and globalParameters["ROCmSMIPath"]:
      runScriptFile.write("%s -d 0 --setfan 255 --setsclk 7\n" % globalParameters["ROCmSMIPath"])
      runScriptFile.write("sleep 1\n")
      runScriptFile.write("%s -d 0 -a\n" % globalParameters["ROCmSMIPath"])

    runScriptFile.write("ERR=0\n")
    for configFile in configPaths:
      runScriptFile.write("{} --config-file {} {} {}\n".format(clientExe, configFile, globalParameters["ClientArgs"], option))
      runScriptFile.write( "if [[ $? -ne 0 ]]\n")
      runScriptFile.write( "then\n")
      runScriptFile.write(f"    echo error in {configFile}\n")
      runScriptFile.write( "    ERR=$?\n")
      runScriptFile.write( "fi\n")

    if globalParameters["PinClocks"] and globalParameters["ROCmSMIPath"]:
      runScriptFile.write("%s -d 0 --resetclocks\n" % globalParameters["ROCmSMIPath"])
      runScriptFile.write("%s -d 0 --setfan 50\n" % globalParameters["ROCmSMIPath"])

    runScriptFile.write("exit $ERR\n")

  runScriptFile.close()

  if os.name != "nt":
    os.chmod(runScriptName, 0o777)

  return runScriptName


def problemSizeParams(problemType, problem):

    numIndices = len(problemType.indices)
    rv = []

    if problem.stridesA:
        astrides = list(problem.stridesA)
    else:
        astrides = [-1] * problemType.aDims
    for sc in problemType.setConstStrideA:
        index = problemType.indices[sc[0]]
        if type(index) == FreeIndex:
            assert(index.isA)
            astrides[index.i] = sc[1]
        else:
            astrides[index.a] = sc[1]

    if problem.stridesB:
      bstrides = list(problem.stridesB)
    else:
      bstrides = [-1] * problemType.bDims
    for sc in problemType.setConstStrideB:
        index = problemType.indices[sc[0]]
        if type(index) == FreeIndex:
            assert(not index.isA)
            bstrides[index.i] = sc[1]
        else:
            bstrides[index.b] = sc[1]

    if problem.stridesC:
      cstrides = list(problem.stridesC)
    else:
      cstrides = [-1] * problemType.cDims

    if problem.stridesD:
      dstrides = list(problem.stridesD)
    else:
      dstrides = [-1] * problemType.dDims

    if len(problem.sizes) == numIndices:
        None
    elif len(problem.sizes) == numIndices + 4:
        # FIXME-problem, this is Exact format with strides tacked onto sizes as 4 extra pams
        # should just set problem.stride* appropriately when reading the Yaml and not deal with extra fields here
        if astrides[1] == -1:
          astrides[1] = problem.sizes[numIndices+2]
        elif astrides[1] != problem.sizes[numIndices+2]:
          raise RuntimeError("problem-specified lda(%u) conflicts with setConstStrideA(%u)" % \
              (astrides[1], problem.sizes[numIndices+2]))

        if bstrides[1] == -1:
          bstrides[1] = problem.sizes[numIndices+3]
        elif bstrides[1] != problem.sizes[numIndices+3]:
          raise RuntimeError("problem-specified ldb(%u) conflicts with setConstStrideB(%u)" % \
              (bstrides[1], problem.sizes[numIndices+3]))

        if cstrides[1] == -1:
          cstrides[1] = problem.sizes[numIndices+1]

        if dstrides[1] == -1:
          dstrides[1] = problem.sizes[numIndices+0]

    else:
        raise RuntimeError(
            "Invalid number of problem type indices: {0} - Indices: {1}, problemSize: {2}".format(len(problem.sizes), numIndices,
            ', '.join(map(str, problem.sizes))))

    problemSizeArg = ('problem-size', ','.join(map(str, problem.sizes[:numIndices])))
    rv.insert(0, problemSizeArg)

    rv.append(('a-strides', ",".join(map(str, astrides))))
    rv.append(('b-strides', ",".join(map(str, bstrides))))
    if cstrides:
      rv.append(('c-strides', ",".join(map(str, cstrides))))
    if dstrides:
      rv.append(('d-strides', ",".join(map(str, dstrides))))

    if problem.zeroPadA:
        rv.append(('a-zero-pads', ';'.join([','.join(map(str,zp)) for zp in problem.zeroPadA])))
    if problem.zeroPadB:
        rv.append(('b-zero-pads', ';'.join([','.join(map(str,zp)) for zp in problem.zeroPadB])))

    return rv

def dataInitParams(problemType):
    initA = globalParameters['DataInitTypeA']
    initB = globalParameters['DataInitTypeB']
    initC = globalParameters['DataInitTypeC']
    initD = globalParameters['DataInitTypeD']
    initAlpha = globalParameters['DataInitTypeAlpha']
    initBeta  = globalParameters['DataInitTypeBeta']

    if not problemType.useBeta:
        initBeta = 0

    if initA == -1: initA = globalParameters['DataInitTypeAB']
    if initB == -1: initB = globalParameters['DataInitTypeAB']

    return [('init-a',     DataInitName(initA).name),
            ('init-b',     DataInitName(initB).name),
            ('init-c',     DataInitName(initC).name),
            ('init-d',     DataInitName(initD).name),
            ('init-alpha', DataInitName(initAlpha).name),
            ('init-beta',  DataInitName(initBeta).name)]

def boundsCheckName(mode):
    if mode == 0: return 'Disable'
    if mode == 1: return 'NaN'
    if mode == 2: return 'GuardPageFront'
    if mode == 3: return 'GuardPageBack'
    if mode == 4: return 'GuardPageAll'


def writeClientConfigIni(problemSizes, problemType, sourceDir, codeObjectFiles, resultsFileName, parametersFilePath, libraryFile=None):

    with open(parametersFilePath, "w") as f:
        def param(key, value):
            f.write("{}={}\n".format(key, value))

        if libraryFile is None:
          libraryFilename = "TensileLibrary.yaml" if globalParameters["LibraryFormat"] == "yaml" else "TensileLibrary.dat"
          libraryFile = os.path.join(sourceDir, "library", libraryFilename)
        param("library-file", libraryFile)

        currentGFXName = Common.gfxName(globalParameters["CurrentISA"])
        for coFile in codeObjectFiles:
            if 'gfx' not in coFile or currentGFXName in coFile:
                param("code-object", os.path.join(sourceDir,coFile))

        param('results-file', resultsFileName)
        convValidation = problemType.convolution and globalParameters["ConvolutionVsContraction"];
        if convValidation:
            param('convolution-identifier', problemType.convolution.identifier())
        param('performance-metric', globalParameters["PerformanceMetric"])
        param('problem-identifier', problemType.operationIdentifier)
        param('a-type',     problemType.aType.toEnum())
        param('b-type',     problemType.bType.toEnum())
        param('c-type',     problemType.cType.toEnum())
        param('d-type',     problemType.dType.toEnum())
        param('alpha-type', problemType.alphaType.toEnum())
        param('beta-type',  problemType.betaType.toEnum())

        param('f32-xdl-math-op', problemType.f32XdlMathOp.toEnum())
        param('stochastic-rounding', problemType.stochasticRounding)

        param('high-precision-accumulate', problemType.highPrecisionAccumulate)
        param('strided-batched', problemType.stridedBatched)

        for problem in problemSizes.problems:
            for key,value in problemSizeParams(problemType, problem):
                param(key,value)
            if convValidation:
              param('convolution-problem', problemType.convolution.identifier(problem))

        param("device-idx",               globalParameters["Device"])

        for key,value in dataInitParams(problemType):
            param(key, value)

        param("c-equal-d",                globalParameters["CEqualD"])

        param("offset-a",                 globalParameters["BufferOffsetA"])
        param("offset-b",                 globalParameters["BufferOffsetB"])
        param("offset-c",                 globalParameters["BufferOffsetC"])
        param("offset-d",                 globalParameters["BufferOffsetD"])

        if globalParameters["PrintTensorA"]:
          param("print-tensor-a",         1)
        if globalParameters["PrintTensorB"]:
          param("print-tensor-b",         1)
        if globalParameters["PrintTensorC"]:
          param("print-tensor-c",         1)
        if globalParameters["PrintTensorD"]:
          param("print-tensor-d",         1)
        if globalParameters["PrintTensorRef"]:
          param("print-tensor-ref",       1)
        if globalParameters["DumpTensors"]:
          param("dump-tensors",           1)
        if globalParameters["ExitOnFails"] > 1:
          param("exit-on-error", 1)

        param("bounds-check",             boundsCheckName(int(globalParameters["BoundsCheck"])))
        param("print-valids",             globalParameters["ValidationPrintValids"])
        param("print-max",                globalParameters["ValidationMaxToPrint"])
        param("num-elements-to-validate", globalParameters["NumElementsToValidate"])
        param("num-benchmarks",           globalParameters["NumBenchmarks"])
        param("num-warmups",              globalParameters["NumWarmups"])
        param("num-enqueues-per-sync",    globalParameters["EnqueuesPerSync"])
        param("num-syncs-per-benchmark",  globalParameters["SyncsPerBenchmark"])
        param("use-gpu-timer",            globalParameters["KernelTime"])
        param("hardware-monitor",         globalParameters["HardwareMonitor"])
        param("flush-count",              globalParameters["FlushCount"])
        param("flush-memory-size",           globalParameters["FlushMemorySize"])
        if convValidation:
            param("convolution-vs-contraction", globalParameters["ConvolutionVsContraction"])
        param("sleep-percent",            globalParameters["SleepPercent"])
        param("perf-l2-read-hits",        globalParameters["PerfModelL2ReadHits"])
        param("perf-l2-write-hits",       globalParameters["PerfModelL2WriteHits"])
        param("perf-l2-read-bw-mul",      globalParameters["PerfModelL2ReadBwMul"])
        param("perf-read-efficiency",     globalParameters["PerfModelReadEfficiency"])
        param("csv-export-extra-cols",    globalParameters["CSVExportWinner"])
        param("csv-merge-same-problems",  globalParameters["CSVMergeSameProblemID"])
        param("log-level",                ClientLogLevel(globalParameters["ClientLogLevel"]).name)
        param("max-workspace-size",       globalParameters["MaxWorkspaceSize"])
        param("granularity-threshold",    globalParameters["GranularityThreshold"])
        param("pristine-on-gpu",          globalParameters["PristineOnGPU"])

        param("library-update-file",      globalParameters["LibraryUpdateFile"])
        param("library-update-comment",   globalParameters["LibraryUpdateComment"])


def writeClientConfig(forBenchmark, solutions, problemSizes, stepName, stepBaseDir, newLibrary, codeObjectFiles, tileAwareSelection, configBase = "ClientParameters", libraryFile = None):

    if tileAwareSelection:
      filename = os.path.join(globalParameters["WorkingPath"], "%s_Granularity.ini"%configBase)
    else:
      filename = os.path.join(globalParameters["WorkingPath"], "%s.ini"%configBase)

    if len(newLibrary.solutions)==0:
      raise RuntimeError ("No valid solutions found")

    resultsFileName = None
    if tileAwareSelection:
      resultsFileName = os.path.join(stepBaseDir, "../Data", stepName+"_Granularity.csv")
    else:
      resultsFileName = os.path.join(stepBaseDir, "../Data", stepName+".csv")

    newSolution = next(iter(newLibrary.solutions.values()))
    sourceDir = os.path.join(stepBaseDir, "source")
    writeClientConfigIni(problemSizes, newSolution.problemType, sourceDir, codeObjectFiles, resultsFileName, filename, libraryFile)

    return filename

def CreateBenchmarkClientParametersForSizes(libraryRootPath, problemSizes, dataFilePath, configFile, problemTypeDict=None):

    libraryPath = os.path.join(libraryRootPath, "library")
    libraryFiles = [os.path.join(libraryPath, f) for f in os.listdir(libraryPath)]
    codeObjectFiles = [f for f in libraryFiles if f.endswith("co")]

    if problemTypeDict:
      problemType = ContractionsProblemType.FromOriginalState(problemTypeDict)
    else:
      # if the we can library contains meta data then we can get the problem type this data
      metaDataFilePath = os.path.join(libraryPath, "metadata.yaml")
      if not os.path.exists(metaDataFilePath):
        printExit ("meta data file %s does not exist" % metaDataFilePath)
      metaData = LibraryIO.readYAML(metaDataFilePath)
      problemTypeDict = metaData["ProblemType"]
      problemType = ContractionsProblemType.FromOriginalState(problemTypeDict)

    writeClientConfigIni(problemSizes, problemType, libraryRootPath, codeObjectFiles, dataFilePath, configFile)

def getClientExecutablePath():
  clientExe = globalParameters.get("PrebuiltClient")

  if not os.path.isfile(clientExe):
    raise FileNotFoundError(
        f"Tensile client executable not found at '{clientExe}'.\n"
        "Please ensure the client is built or provide a valid path using the --prebuilt-client flag.\n"
        "To build, run: `invoke build-client` (you may need to `pip3 install invoke` first).\n"
        "For custom cmake build instructions, please refer to the README in next-cmake."
    )
  return clientExe



################################################################################
#
# Copyright (C) 2020-2022 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

from .Common import globalParameters

import os

class ReplacementKernels:
    def __init__(self, dirpath, codeObjectVersion):
        self.dirpath = dirpath
        self.codeObjectVersion = codeObjectVersion
        self._cache = None

    @property
    def marker(self):
        return '.amdhsa_kernel'

    def getKernelName(self, filename):
        marker = self.marker

        try:
            with open(filename, 'r') as f:
                for line in f:
                    if line.startswith(marker):
                        return line[len(marker):].strip()
        except Exception:
            print(filename)
            raise
        raise RuntimeError("Could not parse kernel name from {}".format(filename))

    @property
    def cache(self):
        if not self._cache:
            self._cache = self.generateCache()
        return self._cache

    def populateCache(self):
        _ = self.cache

    def generateCache(self):
        cache = {}

        for filename in os.listdir(self.dirpath):
            if filename.endswith('.txt'):
                filepath = os.path.join(self.dirpath, filename)
                kernelName = self.getKernelName(filepath)
                if kernelName in cache:
                    raise RuntimeError("Duplicate replacement kernels.  Kernel name: {}, file names: {}, {}".format(kernelName, cache[kernelName], filepath))
                cache[kernelName] = filepath

        return cache

    def get(self, kernelName):
        if kernelName in self.cache:
            return self.cache[kernelName]

        return None

    @classmethod
    def Directory(cls):
        scriptDir = os.path.dirname(os.path.realpath(__file__))
        dirName = 'ReplacementKernels-cov3'

        return os.path.join(scriptDir, dirName)

    _instance = None
    @classmethod
    def Instance(cls):
        if cls._instance:
            return cls._instance
        return cls(cls.Directory(), globalParameters["CodeObjectVersion"])

    @classmethod
    def Get(cls, kernelName):
        return cls.Instance().get(kernelName)



################################################################################
#
# Copyright (C) 2016-2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

def getGlcBitName(memoryModifierFormat):
  if memoryModifierFormat == "GLC":
    return "glc"
  return "sc0"

def getSlcBitName(memoryModifierFormat):
  if memoryModifierFormat == "GLC":
    return "slc"
  return "sc1"

################################################################################
# Memory Instruction
################################################################################
class MemoryInstruction:
  def __init__(self, name, numAddresses, numOffsets, \
      offsetMultiplier, blockWidth, formatting, memoryModifierFormat, forceSC1=False):
    self.name = name
    self.formatting = formatting
    self.numAddresses = numAddresses
    self.numOffsets = numOffsets
    self.offsetMultiplier = offsetMultiplier
    self.blockWidth = blockWidth
    self.memoryModifierFormat = memoryModifierFormat
    self.numBlocks = 2 if self.numAddresses > 1 or self.numOffsets > 1 else 1
    self.totalWidth = self.blockWidth * self.numBlocks
    self.forceSC1 = forceSC1
    #in Quad-Cycle
    if (name == "_ds_load_b128"):
      self.IssueLatency = 2
    elif (name == "_ds_store_b128"):
      self.IssueLatency = 5
    elif (name == "_ds_store2_b64"):
      self.IssueLatency = 3
    elif (name == "_ds_store_b64"):
      self.IssueLatency = 3
    elif (name == "_ds_store2_b32"):
      self.IssueLatency = 3
    elif (name == "_ds_store_b32"):
      self.IssueLatency = 2
    elif (name == "_ds_store_u16") :
      self.IssueLatency = 2
    else:
      self.IssueLatency = 1
    self.endLine = "\n"

  ########################################

  # write in assembly format
  def toString(self, params, comment, nonTemporal=0, highBits=0):
    name = self.name
    if highBits:
      name += "_d16_hi"
    instStr = "%s %s" % (name, (self.formatting % params) )
    if (nonTemporal & 1) != 0 or self.forceSC1:
      instStr += " " + getGlcBitName(self.memoryModifierFormat)
    if (nonTemporal & 2) != 0 or self.forceSC1:
      instStr += " " + getSlcBitName(self.memoryModifierFormat)
    if (nonTemporal & 4) != 0:
      instStr += " nt"
    line = "%-50s // %s%s" % (instStr, comment, self.endLine)
    return line

  # Like toString, but don't add a comment or newline
  # Designed to feed into Code.Inst constructors, somewhat
  def toCodeInst(self, params, nonTemporal=0, highBits=0):
    name = self.name
    if highBits:
      name += "_d16_hi"
    instStr = "%s %s" % (name, (self.formatting % params) )
    if nonTemporal%2==1 or self.forceSC1:
      instStr += " " + getGlcBitName(self.memoryModifierFormat)
    if (nonTemporal//2)%2==1 or self.forceSC1:
      instStr += " " + getSlcBitName(self.memoryModifierFormat)
    if (nonTemporal//4)%2==1:
      instStr += " nt"
    line = "%-50s" % (instStr)
    return line


  def __str__(self):
    return self.name
